import torch
from torch.utils.data import Dataset
import pandas as pd
from datasets import Dataset
from smolhub.helper.dataset.load_config import Config
import torch
from torch.nn.utils.rnn import pad_sequence
from transformers import BatchEncoding


config = Config().get_config()

def json_to_csv(json_dataset):
    # Convert JSON dataset to pd.DataFrame
    csv_dataset = pd.DataFrame(json_dataset)
    return csv_dataset

def _load_dataset(dataset_path):
    # Load dataset from the given path and convert it to a HF Dataset
    if(dataset_path.endswith(".csv")):
        dataset = pd.read_csv(dataset_path)
        hf_dataset = Dataset.from_pandas(dataset)
        return hf_dataset
    elif(dataset_path.endswith(".json")):
        dataset = pd.read_json(dataset_path)
        csv_dataset = json_to_csv(dataset)
        hf_dataset = Dataset.from_pandas(csv_dataset)
        return hf_dataset
    elif(dataset_path.endswith(".xlsx")):
        dataset = pd.read_excel(dataset_path)
        hf_dataset = Dataset.from_pandas(dataset)
        return hf_dataset
    else:
        raise ValueError("Unsupported dataset format")


def dpo_collate_fn_merged_prompt(batch, tokenizer, device):

    merged_chosen_prompts = []
    merged_rejected_prompts = []

    for sample in batch:

        # print(sample)

        # Extract and merge chosen response
        prompt = sample['prompt']
        chosen_data = sample['chosen']
        chosen_data = "Instruction: " + prompt + "\n" + "Output: " + chosen_data[1]['content'] + "\n"
        # Extract and merge rejected response
        rejected_data = sample['rejected']
        rejected_data =  "Instruction: " + prompt + "\n" + "Output: " + rejected_data[1]['content'] + "\n"

        # print(chosen_data)
        # print(rejected_data)
        merged_chosen_prompts.append(chosen_data)


        merged_rejected_prompts.append(rejected_data)

    tokenized_win_prompt = tokenizer(merged_chosen_prompts, max_length = 1024, padding='max_length', truncation=True, return_tensors="pt").to(device)

    tokenized_lose_prompt = tokenizer(merged_rejected_prompts, max_length = 1024, truncation=True, padding='max_length', return_tensors="pt").to(device)



    return {
        # 'prompt': prompts, # Still return original prompts for potential use
        'chosen': tokenized_win_prompt, # List of merged prompt-chosen texts
        'rejected': tokenized_lose_prompt # List of merged prompt-rejected texts
    }

def convert_dataset_to_dataloader_preference(dataset, tokenizer, device):
    # Convert HF Dataset to PyTorch DataLoader

    dataloader = torch.utils.data.DataLoader(dataset, batch_size=config["Dataset"]["batch_size"], num_workers=config["Dataset"]["num_workers"], shuffle=config["Dataset"]["shuffle"], drop_last=config["Dataset"]["drop_last"], pin_memory=config["Dataset"]["pin_memory"], persistent_workers=config["Dataset"]["persistent_workers"], collate_fn = lambda batch: dpo_collate_fn_merged_prompt(batch, tokenizer, device))
  
    return dataloader

class PreferenceDataset(Dataset):
    def __init__(self, dataset_path):
        self.dataset_path = dataset_path
        self.dataset = _load_dataset(dataset_path)
    

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]















from smolhub.scripts.finetune import SFTTrainer
from smolhub.scripts.preference import PreferenceAlignmentTrainer
from smolhub.scripts.preference import PreferenceAlignmentTrainer
from smolhub.scripts.lora import LoRAModel
from transformers import AutoModelForCausalLM, AutoTokenizer
from smolhub.helper.dataset.dataset_main import PreprocessDataset
from smolhub.helper.scheduler import CustomLRScheduler
# from load_config import Config
import torch
from smolhub.helper.dataset.load_config import Config

model_id = "openai-community/gpt2"

config = Config().get_config()
dataset_path = config["Dataset"]["dataset_path"]


sft_model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen2-0.5B-Instruct", token = config["huggingface"]["hf_token"])
ref_model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen2-0.5B-Instruct", token = config["huggingface"]["hf_token"])
tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2-0.5B-Instruct")

# if tokenizer.pad_token is None:
#     # Set the pad token to the eos token if it doesn't exist
#     tokenizer.add_special_tokens({'pad_token': '[PAD]'})
#     # tokenizer.pad_token = tokenizer.eos_token
    

#     print("Setting pad token as PAD token ")

# model.resize_token_embeddings(len(tokenizer))

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


# lora_model = LoRAModel(model)

optimizer = torch.optim.Adam(sft_model.parameters(), lr=1e-6)
scheduler = CustomLRScheduler(optimizer, warmup_iters=100, lr_decay_iters=2000, min_lr=1e-6, max_lr=2e-3, _type="cosine")


#Loading the dataset
preprocess_dataset = PreprocessDataset(dataset_path=dataset_path, tokenizer=tokenizer, device=device)
train_dataloader, val_dataloader, test_dataloader = preprocess_dataset.prepare_dataset()

#Initialize the Trainer
pref_trainer = PreferenceAlignmentTrainer(device, ref_model, sft_model, train_dataloader, val_dataloader, test_dataloader, optimizer, None, tokenizer, scheduler)

#Train
pref_trainer.train()


import os
from pathlib import Path
import torch
from smolhub.helper.dataset.load_config import Config


#Saves the trained model

config = Config().get_config()

class SaveModel:
    def __init__(self, savepath):
        # self.model = model
        self.path = Path.cwd() / savepath  # Corrected to use Path.cwd() for proper path handling
        self.path = str(self.path) 
        self.model_name = config["Model"]["saved_model_name"]
        os.makedirs(self.path, exist_ok=True)
        # self.tokenizer = tokenizer
        
    def save(self, model, tokenizer):
        torch.save(model.state_dict(), os.path.join(self.path, self.model_name))
        torch.save(tokenizer, os.path.join(self.path, 'tokenizer.pt'))
        


import torch
from smolhub.helper.save_model import SaveModel
from smolhub.helper.dataset.load_config import Config

from torch.cuda.amp import GradScaler
from smolhub.helper.count_parameters import count_parameters
from tqdm import tqdm
from smolhub.helper.visualize import Visualizer
# from tests.print_model_details import print_model
import torch.nn as nn

class DPO:
        def __init__(self, ref_model, sft_model, device, beta, tokenizer):


            self.ref_model = ref_model
            self.sft_model = sft_model
            self.device=device
            self.beta = beta
            self.tokenizer = tokenizer
            self.ref_model.eval()



        def DPOloss(self, datapoint):

            self.win_prompt = datapoint['chosen']
            self.lose_prompt = datapoint['rejected']

        #Token level aggregation 
            with torch.no_grad():
                self.win_log_ref = torch.nn.functional.log_softmax(self.ref_model(**self.win_prompt).logits, dim=-1)
                self.win_log_ref = torch.gather(self.win_log_ref, -1, self.win_prompt['input_ids'].unsqueeze(-1)).squeeze(-1) #Why gather? Because its not token level stuff we care about but sequence level. Hence, we will sum up the probs of every token to get seq level but we don't want to do it for attention maksed tokens too. Hence we we will use gather() to get the ids and multiply the probs by the masked out tokens indexes.
                # print("Gather: ", self.chosen_log_probs)
                self.win_log_ref = self.win_log_ref * (self.win_prompt['attention_mask'])
                self.win_log_ref = self.win_log_ref.sum(dim=-1)
                
                self.lose_log_ref = torch.nn.functional.log_softmax(self.ref_model(**self.lose_prompt).logits, dim=-1)
                self.lose_log_ref = torch.gather(self.lose_log_ref, -1, self.lose_prompt['input_ids'].unsqueeze(-1)).squeeze(-1) #Why gather? Because its not token level stuff we care about but sequence level. Hence, we will sum up the probs of every token to get seq level but we don't want to do it for attention maksed tokens too. Hence we we will use gather() to get the ids and multiply the probs by the masked out tokens indexes.
                # print("Gather: ", self.chosen_log_probs)
                self.lose_log_ref = self.lose_log_ref * (self.lose_prompt['attention_mask'])
                self.lose_log_ref = self.lose_log_ref.sum(dim=-1)
            
            self.win_log_sft = torch.nn.functional.log_softmax(self.sft_model(**self.win_prompt).logits, dim=-1)
            self.win_log_sft = torch.gather(self.win_log_sft, -1, self.win_prompt['input_ids'].unsqueeze(-1)).squeeze(-1) #Why gather? Because its not token level stuff we care about but sequence level. Hence, we will sum up the probs of every token to get seq level but we don't want to do it for attention maksed tokens too. Hence we we will use gather() to get the ids and multiply the probs by the masked out tokens indexes.
            self.win_log_sft = self.win_log_sft * (self.win_prompt['attention_mask'])
            self.win_log_sft = self.win_log_sft.sum(dim=-1)
            
            self.lose_log_sft = torch.nn.functional.log_softmax(self.sft_model(**self.lose_prompt).logits, dim=-1)
            self.lose_log_sft = torch.gather(self.lose_log_sft, -1, self.lose_prompt['input_ids'].unsqueeze(-1)).squeeze(-1) #Why gather? Because its not token level stuff we care about but sequence level. Hence, we will sum up the probs of every token to get seq level but we don't want to do it for attention maksed tokens too. Hence we we will use gather() to get the ids and multiply the probs by the masked out tokens indexes.
            self.lose_log_sft = self.lose_log_sft * (self.lose_prompt['attention_mask'])
            self.lose_log_sft = self.lose_log_sft.sum(dim=-1)

            self.diff1 = self.win_log_sft - self.win_log_ref
            self.diff2 = self.lose_log_sft - self.lose_log_ref

            self.final = -nn.functional.logsigmoid(self.beta *(self.diff1 - self.diff2)).mean()

            # sft_model.train()
            return self.final
        
class PreferenceAlignmentTrainer:
    def __init__(self, device, ref_model, sft_model, train_dataloader, val_dataloader, test_dataloader, optimizer, loss_fn, tokenizer, scheduler=None):

        self.ref_model = ref_model
        self.sft_model = sft_model
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.test_dataloader = test_dataloader
        self.device = device
        self.tokenizer = tokenizer
        self.config = Config().get_config()
        self.visualizer = Visualizer()  

        # self.device = 'cpu'
        self.ref_model.to(self.device)
        self.sft_model.to(self.device)
        self.optimizer = optimizer
        self.loss_fn = loss_fn
        self.epochs =  self.config["Model"]["epochs"]
       
        self.eval_iters =  self.config["Model"]["eval_iters"]
        
        #None cus there is loss mask to multiply it with
        if(self.config['Training']['type'] == 'DPO'):
            self.loss_fn = DPO(ref_model, sft_model, device, self.config["Training"]["beta"], tokenizer)
        
        # Mixed Precision Training
        self.scaler = GradScaler(enabled=( self.config["MAP"]["use_float16"]))
        
        #Using torch.compile for efficient training
        if( self.config["Optimizations"]["use_compile"]):
            self.sft_model = torch.compile(self.sft_model)
            self.ref_model = torch.compile(self.ref_model)
        #load scheduler
        self.scheduler = scheduler
        
        self.save_model = SaveModel(self.config['Model']['save_model_path'])
    
    
    


  
    @torch.inference_mode()
    def evaluate(self):
        
        out = {}
        self.sft_model.eval()

        total_loss = 0.0
        total_batches = 0
        
        
        for split in ['val']:
            eval_steps = None
            val_data_iterator = iter(self.val_dataloader)
            eval_steps = self.config["Model"]["eval_steps"]
            
            if eval_steps is None or eval_steps == 0:
                # If eval_steps is not set, use the length of the dataloader
                eval_steps = len(self.val_dataloader)
                
            for k in range(eval_steps):  # Updated to use eval_steps instead of len(self.val_dataloader)
                
                try:
                    batch = next(val_data_iterator)
                except StopIteration:
                    print(f"Resetting val iterator for step {k}")
                    val_data_iterator = iter(self.val_dataloader)
                    batch = next(val_data_iterator)
                
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16 if  self.config["MAP"]["use_bfloat16"] or  self.config["MAP"]["use_float16"] else torch.float32):
                    # batch= next(self.train_dataloader)
                 
                    loss = self.loss_fn.DPOloss(batch)
                    
                total_loss += loss.item()
                total_batches += 1
            
        out[split] = total_loss / total_batches  # Average loss
        

        self.sft_model.train()
        return out

    def train(self):
        
        # self.model.train()
        for epoch in range(self.epochs):
            
            train_iterator = iter(self.train_dataloader)
            for step in tqdm(range(len(self.train_dataloader))):
                
                self.total_steps = len(self.train_dataloader) * self.epochs
                if (step  % self.eval_iters == 0 and step != 0) or step == self.total_steps - 1:
                    losses = self.evaluate()
                    print(f"epoch {epoch}, step {step}: val loss {losses['val']:.4f}")
                    self.visualizer.log({
                        "val_loss": losses['val'],
                        
                    })
                    
                try:
                    batch = next(train_iterator)
                except StopIteration:
                    print(f"Resetting train iterator for epoch {epoch + 1}: step {step}")
                    train_iterator = iter(self.train_dataloader)
                    batch = next(train_iterator)
                
                self.optimizer.zero_grad(set_to_none=True)
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16 if  self.config["MAP"]["use_bfloat16"] or  self.config["MAP"]["use_float16"] else torch.float16):
                  
                    loss = self.loss_fn.DPOloss(batch)
                
                print('Train: ', loss.item())
                loss.requires_grad_(True)
                if  self.config["MAP"]["use_float16"]:
                    self.scaler.scale(loss).backward()
                    # Unscale gradients and optionally clip them
                    # self.scaler.unscale_(self.optimizer)
                  
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                else:
                    loss.backward()
                   
                    self.optimizer.step()
                    # self.optimizer.step()
                self.scheduler.step()

                self.visualizer.log({
                    "epoch": epoch,
                    "step": step,
                    "train_loss": loss.item(),
                    "lr": self.scheduler.get_last_lr()[0],
                    # 'grad_norm': total_norm_before.item(),
                })
                
        print("Training completed!")
        self.visualizer.close()
        
        # Save the model
        self.save_model.save(self.sft_model, self.tokenizer)
        print("Model saved successfully!")